#include "ax.h"

axWindow::axWindow(char *name, int width, int height)
	: name_(name), width_(width), height_(height)
{
	drawable_ = &window_;
	displayname_ = NULL;
	done_ = false;
}

axWindow::~axWindow()
{
	trace("axWindow::~axWindow");
	XFreeGC(display_, gc_);
	XDestroyWindow(display_, window_);
	XCloseDisplay(display_);
}

void axWindow::init()
{
	//XGCValues gcv;
	//XSetWindowAttributes xswa;
	trace("axWindow::init");
	XSizeHints sh;
	unsigned long bgcolor, bdcolor;

	if((display_ = XOpenDisplay(displayname_)) == NULL)
		panic("failed to open display");

	screen_ = DefaultScreen(display_);
	sh.x = 0;
	sh.y = 0;
	sh.width = width_;
	sh.height = height_;
	sh.flags = USPosition | PSize;
	bdcolor = BlackPixel(display_, screen_);
	bgcolor = WhitePixel(display_, screen_);

	window_ = XCreateSimpleWindow(
		display_,
		RootWindow(display_, screen_),
		sh.x, sh.y,
		sh.width, sh.height,
		2, bdcolor, bgcolor);

	XSetWindowBackground(display_, window_, bgcolor);
	XClearWindow(display_, window_);
	XSetStandardProperties(display_, window_, name_, name_, None, NULL, 0, &sh);
	gc_ = XCreateGC(display_, window_, 0, 0);

	// default events
	// NOTE a message map would be more efficient: for each event, this list is
	// walked and all callbacks get called (callBack checks for the right event
	// type before really calling the callback).
	addEvent(new axEvent(this, Expose, &axWindow::onExpose));
	addEvent(new axEvent(this, KeyPress, &axWindow::onKeyPress));
	addEvent(new axEvent(this, KeyRelease, &axWindow::onKeyRelease));
	addEvent(new axEvent(this, ButtonPress, &axWindow::onButtonPress));
	addEvent(new axEvent(this, ButtonRelease, &axWindow::onButtonRelease));

	for(int i = 0; i < events_.size(); i++)
		selectEvents(events_[i].mask_);

	// XXX might move these to run to allow derived classes to do some more
	// stuff after init() before mapping the window
	map();
	flush();
}

void axWindow::run()
{
	trace("axWindow::run");
	init();
	eventLoop();
}

void axWindow::selectEvents(long mask)
{
	XWindowAttributes xwa;
	XSetWindowAttributes xswa;

	if(XGetWindowAttributes(display_, window_, &xwa) != 0){
		xswa.event_mask = xwa.your_event_mask | mask;
		XChangeWindowAttributes(display_, window_, CWEventMask, &xswa);
	}
}

void axWindow::ignoreEvents(long mask)
{
	XWindowAttributes xwa;
	XSetWindowAttributes xswa;

	if(XGetWindowAttributes(display_, window_, &xwa) != 0){
		xswa.event_mask = xwa.your_event_mask & mask;
		XChangeWindowAttributes(display_, window_, CWEventMask, &xswa);
	}
}

void axWindow::checkEvents()
{
	XEvent event;
	
	// ORIG while(XEventsQueued(display_, QueuedAfterReading)){
	while(XPending(display_)){
		XNextEvent(display_, &event);
		// call all of the Event's call back functions to process this event
		for(int i = 0; i < events_.size(); i++)
			events_[i].callBack(event);
	}
}

void axWindow::eventLoop()
{
	XEvent event;

	while(!done_){
		XNextEvent(display_, &event);
		// call all of the Event's call back functions to process this event
		for(int i = 0; i < events_.size(); i++)
			events_[i].callBack(event);
	}
}

void axWindow::addEvent(axEvent *event)
{
	events_.add(event);

	/* auto-add stuff
  XWindowAttributes wa;
  XSetWindowAttributes swa;
  if(XGetWindowAttributes(display_, window_, &wa) != 0){
    swa.event_mask = wa.your_event_mask | event->mask_;
    XChangeWindowAttributes(display_, window_, CWEventMask, &swa);
  }
	*/
}

// default event handlers
void axWindow::onExpose(XEvent &event)
{
	XWindowAttributes wa;

	if(event.xexpose.count == 0){
		trace("axWindow::onExpose");
		XGetWindowAttributes(display_, window_, &wa);
		width_ = wa.width;
		height_ = wa.height;
		drawLine(0, 0, width_, height_);
		drawLine(width_, 0, 0, height_);
		drawString(40, 30, "axtest | press 'q' to quit");
	}
}

void axWindow::onKeyPress(XEvent &event)
{
	KeySym ksym;

	trace("axWindow::onKeyPress");
	ksym = XLookupKeysym((XKeyEvent*)&event, 0);
	switch(ksym){ // see X11/keysymdef.h for keys
		case XK_q:
			done(true);
	}
}

void axWindow::onKeyRelease(XEvent &event)
{
	trace("axWindow::onKeyRelease");
	(void)event;
}

void axWindow::onButtonPress(XEvent &event)
{
	trace("axWindow::onButtonPress");
	(void)event;
}

void axWindow::onButtonRelease(XEvent &event)
{
	trace("axWindow::onButtonRelease");
	(void)event;
}

// X wrappers
inline void axWindow::map()
{
	XMapWindow(display_, window_);
}

inline void axWindow::unmap()
{
	XUnmapWindow(display_, window_);
}

inline void axWindow::flush()
{
	XFlush(display_);
}

inline void axWindow::clearWindow()
{
	XClearWindow(display_, window_);
}

inline void axWindow::getWindowAttributes(XWindowAttributes *wa)
{
	XGetWindowAttributes(display_, window_, wa);
}

inline void axWindow::setForeground(unsigned long pixelvalue)
{
	XSetForeground(display_, gc_, pixelvalue);
}

inline void axWindow::setBackground(unsigned long pixelvalue)
{
	XSetBackground(display_, gc_, pixelvalue);
}

// X wrappers that operate on a drawable
inline void axWindow::drawPoint(int x, int y)
{
	XDrawPoint(display_, *drawable_, gc_, x, y);
}

inline void axWindow::drawLine(int x1, int y1, int x2, int y2)
{
	XDrawLine( display_, *drawable_, gc_, x1, y1, x2, y2);
}

inline void axWindow::drawLines(XPoint *points, int npoints, int mode)
{
	XDrawLines(display_, *drawable_, gc_, points, npoints, mode);
}

inline void axWindow::drawRectangle(int x, int y, uint width, uint height)
{
	XDrawRectangle(display_, *drawable_, gc_, x, y, width, height);
}

inline void axWindow::fillRectangle(int x, int y, uint width, uint height)
{
	XFillRectangle(display_, *drawable_, gc_, x, y, width, height);
	//{	XFillRectangle(display_, window_, gc_, x, y, width + 1, height + 1);
}
inline void axWindow::fillPolygon(XPoint *points, int npoints,
	int shape, int mode)
{
	XFillPolygon(display_, *drawable_, gc_, points, npoints, shape, mode);
}

inline void axWindow::drawString(int x, int y, const char *str)
{
	XDrawString(display_, *drawable_, gc_, x, y, str, strlen(str));
}
