#include "matrix.h"

// OORT matrix coordinates (transposed wrt this code!):
// 0,0 1,0 2,0 3,0
// 0,1 1,1 2,1 3,1
// 0,2 1,2 2,2 3,2
// 0,3 1,3 2,3 3,3

// Coordinate system: conventional right-handed
// +x points right, +y points up, +z points at viewer

#define self (*this) // for readability
const double D2R = M_PI / 180;

Matrix::Matrix() : ROWS(4), COLS(4)
{
	identity();
}

Matrix::~Matrix()
{
}

void Matrix::identity()
{
	for(uint r = 0; r < ROWS; r++)
		for(uint c = 0; c < COLS; c++)
			self(r, c) = (r == c) ? 1.0 : 0.0;
}

Matrix Matrix::identityMatrix()
{
	Matrix t;
	return t;
}

void Matrix::invert()
{
	// Invert the matrix using Gaussian elimination.
	// Not efficient, nor does it check for singular matrices.
	int i, j, k;
	Matrix t;

	if(determinant() == 0.0)
		throw MatrixException("Matrix::invert(): matrix is singular");

	// NOTE this is still in OORT matrix notation, but it doesn't seem to matter
	for(i = 0; i < 4; i++){
		if(self(i, i) != 1.0){
			double divby = self(i, i);
			for(j = 0; j < 4; j++){
				t(i, j)    /= divby;
				self(i, j) /= divby;
			}
		}
		for(j = 0; j < 4; j++){
			if(j != i){
				if(self(j, i) != 0.0){
					double mulby = self(j, i);
					for(k = 0; k < 4; k++){
						self(j, k) -= mulby * self(i, k);
						t(j, k)    -= mulby * t(i, k);
					}
				}
			}
		}
	}
	self = t;
}

Matrix Matrix::invert(const Matrix& m)
{
	Matrix t = m;
	t.invert();
	return t;
}

void Matrix::transpose()
{
	// matrix ^ T
	Matrix t;
	for(uint r = 0; r < ROWS; r++)
		for(uint c = 0; c < COLS; c++)
			t(r, c) = self(c, r);
	self = t;
}

Matrix Matrix::transpose(const Matrix& m)
{
	Matrix t = m;
	t.transpose();
	return t;
}

double Matrix::determinant() const
{
	// 4x4 matrix determinant, generated with Mathematica
	// determinant == 0 -> matrix is singular, no inverse
	// m03 m12 m21 m30 - m02 m13 m21 m30 - m03 m11 m22 m30 + m01 m13 m22 m30 + 
	// m02 m11 m23 m30 - m01 m12 m23 m30 - m03 m12 m20 m31 + m02 m13 m20 m31 + 
	// m03 m10 m22 m31 - m00 m13 m22 m31 - m02 m10 m23 m31 + m00 m12 m23 m31 + 
	// m03 m11 m20 m32 - m01 m13 m20 m32 - m03 m10 m21 m32 + m00 m13 m21 m32 + 
	// m01 m10 m23 m32 - m00 m11 m23 m32 - m02 m11 m20 m33 + m01 m12 m20 m33 + 
	// m02 m10 m21 m33 - m00 m12 m21 m33 - m01 m10 m22 m33 + m00 m11 m22 m33
	return
		self(0,3) * self(1,2) * self(2,1) * self(3,0) -
		self(0,2) * self(1,3) * self(2,1) * self(3,0) -
		self(0,3) * self(1,1) * self(2,2) * self(3,0) +
		self(0,1) * self(1,3) * self(2,2) * self(3,0) +
		self(0,2) * self(1,1) * self(2,3) * self(3,0) -
		self(0,1) * self(1,2) * self(2,3) * self(3,0) -
		self(0,3) * self(1,2) * self(2,0) * self(3,1) +
		self(0,2) * self(1,3) * self(2,0) * self(3,1) +
		self(0,3) * self(1,0) * self(2,2) * self(3,1) -
		self(0,0) * self(1,3) * self(2,2) * self(3,1) -
		self(0,2) * self(1,0) * self(2,3) * self(3,1) +
		self(0,0) * self(1,2) * self(2,3) * self(3,1) +
		self(0,3) * self(1,1) * self(2,0) * self(3,2) -
		self(0,1) * self(1,3) * self(2,0) * self(3,2) -
		self(0,3) * self(1,0) * self(2,1) * self(3,2) +
		self(0,0) * self(1,3) * self(2,1) * self(3,2) +
		self(0,1) * self(1,0) * self(2,3) * self(3,2) -
		self(0,0) * self(1,1) * self(2,3) * self(3,2) -
		self(0,2) * self(1,1) * self(2,0) * self(3,3) +
		self(0,1) * self(1,2) * self(2,0) * self(3,3) +
		self(0,2) * self(1,0) * self(2,1) * self(3,3) -
		self(0,0) * self(1,2) * self(2,1) * self(3,3) -
		self(0,1) * self(1,0) * self(2,2) * self(3,3) +
		self(0,0) * self(1,1) * self(2,2) * self(3,3);
}

double Matrix::determinant(const Matrix& m)
{
	return m.determinant();
}

// create various types of matrices
Matrix Matrix::translateMatrix(const Vector& v)
{
	Matrix t;
	t(0,3) = v.x();
	t(1,3) = v.y();
	t(2,3) = v.z();
	return t;
}

Matrix Matrix::scaleMatrix(double d)
{
	Matrix t;
	t(0,0) = t(1,1) = t(2,2) = d;
	return t;
}

Matrix Matrix::rotateXMatrix(double angle) // pitch
{
	Matrix t;
	double cx = cos(angle * D2R);
	double sx = sin(angle * D2R);
	t(1,1) =  cx;
	t(1,2) = -sx;
	t(2,1) =  sx;
	t(2,2) =  cx;
	return t;
}

Matrix Matrix::rotateYMatrix(double angle) // yaw
{
	Matrix t;
	double cy = cos(angle * D2R);
	double sy = sin(angle * D2R);
	t(0,0) =  cy;
	t(0,2) =  sy;
	t(2,0) = -sy;
	t(2,2) =  cy;
	return t;
}

Matrix Matrix::rotateZMatrix(double angle) // roll
{
	Matrix t;
	double cz = cos(angle * D2R);
	double sz = sin(angle * D2R);
	t(0,0) =  cz;
	t(0,1) = -sz;
	t(1,0) =  sz;
	t(1,1) =  cz;
	return t;
}

Matrix Matrix::rotateXYZMatrix(const Vector& v) // pitch, yaw, roll
{
	Matrix t;
	double cx = cos(v.x() * D2R); // pitch
	double sx = sin(v.x() * D2R);
	double cy = cos(v.y() * D2R); // yaw
	double sy = sin(v.y() * D2R);
	double cz = cos(v.z() * D2R); // roll
	double sz = sin(v.z() * D2R);
	t(0,0) = cz * cy;
	t(0,1) = -sz * cy;
	t(0,2) = sy;
	t(1,0) = sz * cx + cz * sy * sx;
	t(1,1) = cz * cx - sz * sy * sx;
	t(1,2) = -cy * sx;
	t(2,0) = sz * sx - cz * sy * cx;
	t(2,1) = cz * sx + sz * sy * cx ;
	t(2,2) = cy * cx;
	return t;
}

double& Matrix::operator()(uint row, uint col)
{
	// assignments to elements: m(i,j) = x;
	if(row >= ROWS || col >= COLS)
		throw MatrixException("Matrix::operator(): subscript out of bounds");
	return data_[COLS * row + col];
}

double Matrix::operator()(uint row, uint col) const
{
	// constant use of elements, e.g. in operators with const Matrix& params
	if (row >= ROWS || col >= COLS)
		throw MatrixException("Matrix::operator() const: subscript out of bounds");
	return data_[COLS * row + col];
} 

Matrix& Matrix::operator=(const Matrix& m)
{
	// override of default `bit-wise copy' assignment: neccessary due to data_
	// member which needs indirection to be copied
	// old style:
	// memcpy(data_, m.data_, sizeof(double) * ROWS * COLS);
	// XXX casting away const to shut up gcc-2.95.2, hairy!
	data_ = ((Matrix)m).data_;
	return *this;
}

Matrix Matrix::operator*(const Matrix& m) const
{
	// this->operator*(m) <==> this * m
	// matrix = matrix * matrix
	Matrix t;
	for(uint r = 0; r < ROWS; r++)
		for(uint c = 0; c < m.COLS; c++){
			t(r, c) = 0.0;
			for(uint i = 0; i < ROWS; i++)
				t(r, c) += self(r, i) * m(i, c);
		}
	//XXX only in *= += etc.
	//*this = t;
	//return *this;
	return t;
}

Vector Matrix::operator*(const Vector& v) const
{
	// vector = matrix * vector
	Vector t;
	for(uint r = 0; r < ROWS; r++){
		t(r) = 0;
		for(uint c = 0; c < COLS; c++)
			t(r) += self(r, c) * v(c);
	}
	double denominator = self(3,0) + self(3,1) + self(3,2) + self(3,3);
	if(denominator != 1.0)
		t = t / denominator;
	return t;
}

ostream& operator<<(ostream& s, const Matrix& m)
{
	s << "+---\n";
	for(uint r = 0; r < m.ROWS; r++){
		s << "|";
		for(uint c = 0; c < m.COLS; c++)
			s << setprecision(3) << setw(7) << m(r, c);
			//printf(" %6.2f", (*this)(r, c));
		s << "\n";
	}
	return s;
}
